#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <inttypes.h>
#include <errno.h>
#include <signal.h>
#include <assert.h>
#include <fcntl.h>
#include <setjmp.h>
#include <pthread.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <uuid/uuid.h>

#include "msr-index.h"
#include "xenner.h"
#include "mm.h"

/* ------------------------------------------------------------------ */

static int shutdown;
static int termsig;
static int timer;
static int pending_io;

struct stats {
    int sig_alrm;
    int sig_io;
    int cb_out;
    int cb_run;
    int cb_hlt;
} cs, ls;

/* ------------------------------------------------------------------ */

static int shutdown;
static int termsig;
static int timer;
static int pending_io;

static sigset_t wait_suspend;
static sigset_t wait_block;

static void sighandle(int sig)
{
    switch (sig) {
    case SIGHUP:
	/* ignore */
	break;
    case SIGALRM:
	cs.sig_alrm++;
	timer++;
	break;
    case SIGIO:
	cs.sig_io++;
	pending_io++;
	break;
    case SIGINT:
	if (!shutdown) {
	    shutdown++;
	    break;
	}
	/* else fall through */
    case SIGTERM:
    case SIGQUIT:
    default:
	fprintf(stderr, "%s: %s\n", __FUNCTION__, strsignal(sig));
	termsig = sig;
    }
}

static void sigsetup(struct xenvm *xen)
{
    struct sigaction act, old;

    /* install handlers */
    memset(&act,0,sizeof(act));
    act.sa_handler = sighandle;
    sigemptyset(&act.sa_mask);
    sigaction(SIGHUP,  &act, &old);
    sigaction(SIGINT,  &act, &old);
    sigaction(SIGTERM, &act, &old);
    sigaction(SIGQUIT, &act, &old);
    sigaction(SIGALRM, &act, &old);
    sigaction(SIGIO,   &act, &old);

    /* setup masks (see wait_event() below) */
    sigfillset(&wait_suspend);
    sigdelset(&wait_suspend, SIGINT);
    sigdelset(&wait_suspend, SIGTERM);
    sigdelset(&wait_suspend, SIGQUIT);
    sigdelset(&wait_suspend, SIGALRM);
    sigdelset(&wait_suspend, SIGIO);

    sigemptyset(&wait_block);
    sigaddset(&wait_block,   SIGALRM);
    sigaddset(&wait_block,   SIGIO);
}

void alive(void)
{
    static const char alive[] = ".oOo";
    static const int chars = sizeof(alive)/sizeof(alive[0]) -1;
    static int count;
    
    fprintf(stderr, "[%c]\r", alive[count++ % chars]);
}

uint64_t get_systime(void)
{
    uint64_t systime;
    struct timeval tv;

    gettimeofday(&tv, NULL);
    systime  = tv.tv_sec  * (uint64_t)1000000000;
    systime += tv.tv_usec * (uint64_t)1000;
    return systime;
}

static int balloon(struct xenvm *xen)
{
    uint32_t page;
    int advice;

    /* Hmm, keep track of state and signal only changes?  Not sure it is
     * worth the effort given ballooning is an infrequent operation in
     * Xen due to being fully admin-controlled (aka manual) */
    for (page = 0; page < xen->pg_guest; page++) {
	advice = 0;
	if (xen->m2p_32)
	    if (xen->m2p_32[page] == (uint32_t)INVALID_M2P_ENTRY)
		advice = MADV_DONTNEED;
	if (xen->m2p_64)
	    if (xen->m2p_64[page] == (uint64_t)INVALID_M2P_ENTRY)
		advice = MADV_DONTNEED;
	if (advice)
	    madvise(xen->memory + page * PAGE_SIZE, PAGE_SIZE, advice);
    }
    return 0;
}

/* ------------------------------------------------------------------ */

static void systime_to_tv(uint64_t systime, struct timeval *tv)
{
    tv->tv_sec  = systime / 1000000000;
    tv->tv_usec = (systime / 1000) % 1000000;
}

void setup_timer(struct xenvcpu *vcpu)
{
    struct itimerval it;
    int64_t next_tick;

    memset(&it, 0, sizeof(it));
    if (vcpu->period) {
	systime_to_tv(vcpu->period, &it.it_interval);
	systime_to_tv(vcpu->period, &it.it_value);
    }
    if (vcpu->oneshot) {
	uint64_t now = get_systime();
	next_tick = vcpu->oneshot - (now - vcpu->vm->boot);
	if (next_tick <= 0) {
	    timer++;
	    next_tick = vcpu->period;
	    if (0 == next_tick)
		next_tick = XEN_DEFAULT_PERIOD;
	}
	systime_to_tv(next_tick, &it.it_value);
    }
    setitimer(ITIMER_REAL, &it, NULL);
}

static void stop_timer(void)
{
    struct itimerval it;

    memset(&it, 0, sizeof(it));
    setitimer(ITIMER_REAL, &it, NULL);
}

static inline void print_stats(uint64_t now)
{
    static uint64_t last_stats;

    if (now - last_stats < 1000000000)
	return;
    fprintf(stderr,
	    "sig: alrm %4d io %4d  |  "
	    "cb: out %4d, run %4d hlt %4d  |  "
	    "\n",
	    cs.sig_alrm - ls.sig_alrm,
	    cs.sig_io   - ls.sig_io,
	    cs.cb_out   - ls.cb_out,
	    cs.cb_run   - ls.cb_run,
	    cs.cb_hlt   - ls.cb_hlt);
    ls = cs;
    last_stats = now;
}

void update_time(struct xenvcpu *vcpu)
{
    struct xenvm *xen = vcpu->vm;
    uint64_t now, systime;

    now = get_systime();
    if (xen->print_stats && !xen->nostderr)
	print_stats(now);

    if (now - vcpu->systime > 1000000)
	vcpu->update_time = 1;
    if (!vcpu->update_time)
	return;

    vcpu->update_time = 0;
    vcpu->systime = now;
    systime = vcpu->systime - vcpu->vm->boot;

    d2printf("systime %" PRId64 ".%09" PRId64 "\n",
	     systime / 1000000000, systime % 1000000000);

#if 1 /* FIXME: move elsewhere */
    xen->xen_wc->wc_version++;
    xen->xen_wc->wc_sec  = xen->boot / 1000000000;
    xen->xen_wc->wc_nsec = xen->boot % 1000000000;
    xen->xen_wc->wc_version++;
#endif

    if (vcpu->xentime) {
	vcpu->xentime->version++;
	vcpu->xentime->system_time       = systime;
	vcpu->xentime->tsc_timestamp     = get_msr(vcpu, MSR_IA32_TSC);
	vcpu->xentime->tsc_to_system_mul = xen->tsc_mul_frac;
	vcpu->xentime->tsc_shift         = xen->tsc_shift;
	vcpu->xentime->version++;
    }
}

int raise_event(struct xenvcpu *vcpu, int port)
{
    struct xenvm *xen = vcpu->vm;
    uint32_t *p32, *m32, *s32;
    uint64_t *p64, *m64, *s64;

    if (kvm_irqchip_in_kernel(xen->kvm)) {
	kvm_set_irq_level(xen->kvm, port, 1);
	return 0;
    }

    switch (xen->mode) {
    case XENMODE_32:
    case XENMODE_PAE:
	p32 = xen->shinfo.sh32->evtchn_pending;
	m32 = xen->shinfo.sh32->evtchn_mask;
	s32 = &xen->shinfo.sh32->vcpu_info[vcpu->id].evtchn_pending_sel;

	/* flag event port as pending */
	if (TESTBIT(p32, port))
	    break;
	SETBIT(p32, port);

	/* stop if event port is masked */
	if (TESTBIT(m32, port))
	    break;

	/* flag vcpu has events */
	if (TESTBIT(s32, port / 32))
	    break;
	SETBIT(s32, port / 32);
	*(xen->vcpu[vcpu->id].evtchn_upcall_pending) = 1;
	break;
    case XENMODE_64:
	p64 = xen->shinfo.sh64->evtchn_pending;
	m64 = xen->shinfo.sh64->evtchn_mask;
	s64 = &xen->shinfo.sh64->vcpu_info[vcpu->id].evtchn_pending_sel;

	/* flag event port as pending */
	if (TESTBIT(p64, port))
	    break;
	SETBIT(p64, port);

	/* stop if event port is masked */
	if (TESTBIT(m64, port))
	    break;

	/* flag vcpu has events */
	if (TESTBIT(s64, port / 64))
	    break;
	SETBIT(s64, port / 64);
	*(xen->vcpu[vcpu->id].evtchn_upcall_pending) = 1;
	break;
    }
    return 0;
}

#define have_event(_v) (*((_v)->evtchn_upcall_pending))
#define mask_event(_v) (*((_v)->evtchn_upcall_mask))

static int send_event(struct xenvcpu *vcpu)
{
    struct xenvm *xen = vcpu->vm;

    if (kvm_irqchip_in_kernel(xen->kvm))
	vm_kill(vcpu, "send_event: Oops, irqchip", 0);
    return have_event(vcpu) && !mask_event(vcpu);
}

void wait_event(struct xenvcpu *vcpu)
{
    struct xenvm *xen = vcpu->vm;
    sigset_t orig;
    
    if (kvm_irqchip_in_kernel(xen->kvm))
	vm_kill(vcpu, "wait_event: Oops, irqchip", 0);

    if (have_event(vcpu))
	return;
    
    /* balloon: check for work before going sleep */
    if (vcpu->vm->vminfo->balloon) {
	balloon(vcpu->vm);
	vcpu->vm->vminfo->balloon = 0;
    }

    sigprocmask(SIG_BLOCK, &wait_block, &orig);
    if (!timer && !pending_io) {
	pthread_mutex_unlock(&vcpu->vm->biglock);
	sigsuspend(&wait_suspend);
	pthread_mutex_lock(&vcpu->vm->biglock);
	vcpu->update_time = 1;
    }
    sigprocmask(SIG_SETMASK, &orig, NULL);
}

int xenner_register_cb(struct xenvm *xen, enum xenner_cb_type type,
		       char *name, int handle, void *cb_data,
		       int (*callback)(struct xenvm *xen, int fd, void *data))
{
    struct xenner_callback *cb;
    long flags;

    cb = malloc(sizeof(*cb));
    memset(cb,0,sizeof(*cb));
    cb->handle   = handle;
    cb->name     = name;
    cb->cb_data  = cb_data;
    cb->callback = callback;
    list_add_tail(&cb->list, &xen->callbacks[type]);

    switch (type) {
    case XENNER_CB_IOFD:
	/* setup stuff so we get SIGIO on data */
	flags  = fcntl(cb->handle, F_GETFL, 0);
	flags |= O_ASYNC | O_NONBLOCK;
	if (0 != fcntl(cb->handle, F_SETFL, flags)) {
	    perror("fcntl F_SETFL");
	    return -1;
	}
	if (0 != fcntl(cb->handle, F_SETOWN, getpid())) {
	    perror("fcntl F_SETOWN");
	    return -1;
	}
	if (0 != fcntl(cb->handle, F_SETSIG, SIGIO)) {
	    perror("fcntl F_SETSIG");
	    return -1;
	}
	break;
    default:
	/* nothing special to do */
	break;
    }
    return 0;
}

static int callback_iofds(struct xenvm *xen)
{
    struct list_head *item;
    struct xenner_callback *cb;
    int count;

    list_for_each(item, &xen->callbacks[XENNER_CB_IOFD]) {
	cb = list_entry(item, struct xenner_callback, list);
	count = cb->callback(xen, cb->handle, cb->cb_data);
    }
    return 0;
}

int xenner_callback_evtchn(struct xenvm *xen, int port)
{
    struct list_head *item;
    struct xenner_callback *cb;

    list_for_each(item, &xen->callbacks[XENNER_CB_EVTCHN]) {
	cb = list_entry(item, struct xenner_callback, list);
	if (cb->handle != port)
	    continue;
	d2printf("%s: calling %s\n", __FUNCTION__, cb->name);
	cb->callback(xen, cb->handle, cb->cb_data);
	return 0;
    }
    return -1;
}

static void check_events(struct xenvcpu *vcpu)
{
    struct xenvm *xen = vcpu->vm;

    d3printf("%s: io %d timer %d\n", __FUNCTION__, pending_io, timer);

    if (shutdown) {
	static int once = 1;
	if (once) {
	    xenstore_shutdown(xen, "halt");
	    once = 0;
	}
    }

    if (pending_io) {
	pending_io = 0;
	callback_iofds(xen);
    }

    if (timer) {
	timer = 0;
	if (xen->vminfo->evtchn_timer) {
	    raise_event(vcpu, xen->vminfo->evtchn_timer);
	} else if (xen->virq[VIRQ_TIMER]) {
	    raise_event(vcpu, xen->virq[VIRQ_TIMER]);
	}
    }
}

/* ------------------------------------------------------------------ */

void need_regs(struct xenvcpu *vcpu)
{
    if (vcpu->regs_up2date)
	return;
    kvm_get_regs(vcpu->vm->kvm, vcpu->id, &vcpu->regs);
    vcpu->regs_up2date = 1;
}

void need_sregs(struct xenvcpu *vcpu)
{
    if (vcpu->sregs_up2date)
	return;
    kvm_get_sregs(vcpu->vm->kvm, vcpu->id, &vcpu->sregs);
    vcpu->sregs_up2date = 1;
}

void flush_regs(struct xenvcpu *vcpu)
{
    assert(vcpu->regs_up2date);
    vcpu->regs_dirty = 1;
}

void flush_sregs(struct xenvcpu *vcpu)
{
    assert(vcpu->sregs_up2date);
    vcpu->sregs_dirty = 1;
}

void xenner_cleanup(struct xenvm *xen, int stats)
{
    /* going down now ... */
    if (xen->vminfo)
	xen->vminfo->dying = 1;

    if (stats) {
	section_print(xen, __FUNCTION__, "statistics");
	hypercall_stats(xen);
    }

    section_print(xen, __FUNCTION__, "cleaning up");
    xenstore_fini(xen);
    evtchn_fini(xen);
    qemu_cleanup(xen);
    stop_timer();
    vmcore_fini(xen);
}

/* ------------------------------------------------------------------ */

static void print_stack_32(char *name, uint32_t esp, uint32_t *ptr)
{
    int i;
    
    fprintf(stderr, "--- %s stack trace ---\n", name);
    for (i = 0; i < 128; i++) {
	if (0 == addr_offset(esp + 4*i))
	    break;
	if (0 == (i % 4))
	    fprintf(stderr, "%08" PRIx32 ":", esp + 4*i);
	fprintf(stderr, " %08" PRIx32, ptr[i]);
	if (3 == (i % 4))
	    fprintf(stderr, "\n");
    }
    if (0 != (i % 4))
	fprintf(stderr, "\n");
}

static void print_stack_64(char *name, uint64_t esp, uint64_t *ptr)
{
    int i;
    
    fprintf(stderr, "--- %s stack trace ---\n", name);
    for (i = 0; i < 128; i++) {
	if (0 == addr_offset(esp + 8*i))
	    break;
	if (0 == (i % 4))
	    fprintf(stderr, "%016" PRIx64 ":", esp + 8*i);
	fprintf(stderr, " %016" PRIx64, ptr[i]);
	if (3 == (i % 4))
	    fprintf(stderr, "\n");
    }
    if (0 != (i % 4))
	fprintf(stderr, "\n");
}

void print_guest_stack(struct xenvcpu *vcpu, uint64_t rsp)
{
    if (vcpu->vm->mode == XENMODE_64) {
	uint64_t *data = guest_vaddr_to_ptr(vcpu, rsp);
	if (data)
	    print_stack_64("guest64", rsp, data);
    } else {
	uint32_t *data = guest_vaddr_to_ptr(vcpu, rsp);
	if (data)
	    print_stack_32("guest32", rsp, data);
    }
}

void print_emu_stack(struct xenvm *xen, uint64_t rsp)
{
    if (xen->mode == XENMODE_64) {
	uint64_t *data = emu_vaddr_to_ptr(xen, rsp);
	if (data)
	    print_stack_64("emu64", rsp, data);
    } else {
	uint32_t *data = emu_vaddr_to_ptr(xen, rsp);
	if (data)
	    print_stack_32("emu32", rsp, data);
    }
}

void vm_kill(struct xenvcpu *vcpu, char *reason, uint64_t guest_rsp)
{
    struct xenvm *xen = vcpu->vm;
    
    banner_print(xen, reason);
    fprintf(stderr, "--- registers (vcpu %d)---\n", vcpu->id);
    kvm_show_regs(xen->kvm, vcpu->id);
    print_msrs(vcpu);

    need_regs(vcpu);
    if (vcpu->regs.rsp > xen->emu_vs && vcpu->regs.rsp < xen->emu_ve)
	print_emu_stack(xen, vcpu->regs.rsp);
    else if (!guest_rsp && vcpu->regs.rsp)
	guest_rsp = vcpu->regs.rsp;

    if (guest_rsp)
	print_guest_stack(vcpu, guest_rsp);	

    xenner_cleanup(xen, 1);
    exit(1);
}

/* ------------------------------------------------------------------ */
/* functions                                                          */

static int cb_inb(void *opaque, uint16_t addr, uint8_t *value)
{
    struct xenvm *xen = opaque;

    d2printf("io: %s, addr 0x%" PRIx16 "\n",
	     __FUNCTION__, addr);
    *value = 0xff;
    return 0;
}

static int cb_inw(void *opaque, uint16_t addr, uint16_t *value)
{
    struct xenvm *xen = opaque;

    d2printf("io: %s, addr 0x%" PRIx16 "\n",
	     __FUNCTION__, addr);
    *value = 0xffff;
    return 0;
}

static int cb_inl(void *opaque, uint16_t addr, uint32_t *value)
{
    struct xenvm *xen = opaque;

    d2printf("io: %s, addr 0x%" PRIx16 "\n",
	     __FUNCTION__, addr);
    *value = 0xffffffff;
    return 0;
}

static int cb_outb(void *opaque, uint16_t addr, uint8_t value, int vcpu_id)
{
    struct xenvm *xen = opaque;
    struct xenvcpu *vcpu = xen->vcpu + vcpu_id;
    char *msg;

    cs.cb_out++;
    switch (addr) {
#if 1 /* handled by emu now, shouldn't happen any more */
    case 0xe1: /* illegal instruction */
	vm_kill(vcpu, "illegal instruction (obsolete)", 0);
	// do_instr_emu(xen);
	break;
    case 0xe2: /* page fault */
	vm_kill(vcpu, "page fault (obsolete)", 0);
	break;
    case 0xe3: /* general protection */
	vm_kill(vcpu, "general protection (obsolete)", 0);
//	do_general_protection(xen);
	break;
    case 0xe4: /* division by zero */
	vm_kill(vcpu, "division by zero (obsolete)", 0);
	break;
    case 0xe5: /* event callback (irq 0x20) */
	vm_kill(vcpu, "event callback (obsolete)", 0);
	break;
#endif

    case 0xe0: /* hypercall */
	do_xen_hypercall(vcpu, value);
	break;

    case 0xea: /* emu: print string */
	need_regs(vcpu);
	msg = emu_vaddr_to_ptr(xen, vcpu->regs.rdx);
	conemu_print(vcpu, msg);
	break;
    case 0xeb: /* emu: panic */
	vm_kill(vcpu, "emu: panic()", 0);
	break;

    case 0xec: /* debug trap */
	do_debug_trap(vcpu);
	break;

    case 0xed: /* shutdown */
	logprintf(xen, "guest requests shutdown, exiting.\n");
	xenner_cleanup(xen, 1);
	exit(0);

    case 0xee: /* get evtchn port */
	need_regs(vcpu);
	vcpu->regs.rax = evtchn_port(xen, "irqchip");
	flush_regs(vcpu);
	break;

    case 0xef: /* unhandled interrupt */
	vm_kill(vcpu, "unknown irq", 0);
	break;
    default:
	d2printf("io: %s, addr 0x%" PRIx16 " value 0x%" PRIx8 "\n",
		__FUNCTION__, addr, value);
	break;
    }
    return 0;
}

static int cb_outw(void *opaque, uint16_t addr, uint16_t value)
{
    struct xenvm *xen = opaque;

    d1printf("io: %s, addr 0x%" PRIx16 " value 0x%" PRIx16 "\n",
	     __FUNCTION__, addr, value);
    return 0;
}

static int cb_outl(void *opaque, uint16_t addr, uint32_t value)
{
    struct xenvm *xen = opaque;

    d1printf("io: %s, addr 0x%" PRIx16 " value 0x%" PRIx32 "\n",
	     __FUNCTION__, addr, value);
    return 0;
}

static int cb_debug(void *opaque, int vcpu_id)
{
    struct xenvm *xen = opaque;
    
    d3printf("callback: %s/%d\n", __FUNCTION__, vcpu_id);
    return 0;
}

static int cb_halt(void *opaque, int vcpu_id)
{
    struct xenvm *xen = opaque;
    struct xenvcpu *vcpu = xen->vcpu + vcpu_id;

    cs.cb_hlt++;
    d3printf("callback: %s/%d\n", __FUNCTION__, vcpu_id);
    wait_event(vcpu);
    return 0;
}

static int cb_shutdown(void *opaque, int vcpu_id)
{
    struct xenvm *xen = opaque;
    struct xenvcpu *vcpu = xen->vcpu + vcpu_id;

    d3printf("callback: %s/%d\n", __FUNCTION__, vcpu_id);
    vm_kill(vcpu, "crashed", 0);
    return 0;
}

static int cb_io_window(void *opaque)
{
    struct xenvm *xen = opaque;

    d2printf("callback: %s\n", __FUNCTION__);
    return 0;
}

static int cb_try_push_interrupts(void *opaque)
{
    struct xenvm *xen = opaque;
    struct xenvcpu *vcpu = xen->vcpu + 0; /* FIXME */
    
    d3printf("callback: %s\n", __FUNCTION__);

    check_events(vcpu);

    if (send_event(vcpu))
	kvm_inject_irq(xen->kvm, vcpu->id, 0x20);

    return 0;
}

static void cb_post_kvm_run(void *opaque, int vcpu_id)
{
    struct xenvm *xen = opaque;

    pthread_mutex_lock(&xen->biglock);
    d3printf("callback: %s/%d\n", __FUNCTION__, vcpu_id);
}

static int cb_pre_kvm_run(void *opaque, int vcpu_id)
{
    struct xenvm *xen = opaque;
    struct xenvcpu *vcpu = xen->vcpu + vcpu_id;

    cs.cb_run++;
    d3printf("callback: %s/%d\n", __FUNCTION__, vcpu_id);
    if (termsig)
	vm_kill(vcpu, "termsig", 0);

    if (kvm_irqchip_in_kernel(xen->kvm))
	check_events(vcpu);

    /* timekeeping */
    update_time(vcpu);

    /* registers */
    if (vcpu->regs_dirty) {
	kvm_set_regs(xen->kvm, vcpu->id, &vcpu->regs);
	vcpu->regs_dirty = 0;
    }
    if (vcpu->sregs_dirty) {
	kvm_set_sregs(xen->kvm, vcpu->id, &vcpu->sregs);
	vcpu->sregs_dirty = 0;
    }
    vcpu->regs_up2date  = 0;
    vcpu->sregs_up2date = 0;

    pthread_mutex_unlock(&xen->biglock);
    return 0;
}

static void *vcpu_thread(void *arg)
{
    struct xenvcpu *vcpu = arg;
    struct xenvm *xen = vcpu->vm;

    pthread_mutex_lock(&xen->biglock);
    d1printf("%s/%d: start\n", __FUNCTION__, vcpu->id);
    vcpu->online = 1;

    kvm_run(xen->kvm, vcpu->id);

    vcpu->online = 0;
    d1printf("%s/%d: stop\n", __FUNCTION__, vcpu->id);
    pthread_mutex_unlock(&xen->biglock);
    return NULL;
}

/* ------------------------------------------------------------------ */

static struct kvm_callbacks xenner_cb = {
    .inb                 = cb_inb,
    .inw                 = cb_inw,
    .inl                 = cb_inl,
    .outb                = cb_outb,
    .outw                = cb_outw,
    .outl                = cb_outl,
    .debug               = cb_debug,
    .halt                = cb_halt,
    .shutdown            = cb_shutdown,
    .io_window           = cb_io_window,
    .try_push_interrupts = cb_try_push_interrupts,
    .post_kvm_run        = cb_post_kvm_run,
    .pre_kvm_run         = cb_pre_kvm_run,
};

int kvmbits_init(struct xenvm *xen, int irqchip)
{
    int i;
    
    /* init kvm */
    xen->kvm = kvm_init(&xenner_cb, xen);
    if (!xen->kvm) {
	fprintf(stderr, "kvm_init failed\n");
	goto err;
    }
    
    /* create vm */
    xen->pg_total &= ~((1<<10)-1);  /* round to 4MB border */
    if (vmcore_init(xen) < 0)
	goto err;
    if (kvm_create_vm(xen->kvm) < 0) {
	fprintf(stderr, "kvm_create_vm failed\n");
	goto err;
    }

#ifndef KVM_CAP_SET_TSS_ADDR
#error linux/kvm.h is too old (need 49+).
#endif
    if (kvm_set_tss_addr(xen->kvm, xen->pg_total * PAGE_SIZE) < 0) {
	fprintf(stderr, "kvm_set_tss_addr failed (%s)\n", strerror(errno));
	fprintf(stderr, "Old kvm version maybe?  I need 49+.\n");
	goto err;
    }

    if (kvm_register_userspace_phys_mem(xen->kvm, 0, xen->memory,
					xen->pg_total * PAGE_SIZE, 1) < 0) {
	fprintf(stderr, "kvm_register_userspace_phys_mem failed\n");
	goto err;
    }

    if (irqchip) {
	kvm_create_irqchip(xen->kvm);
	if (!kvm_irqchip_in_kernel(xen->kvm)) {
	    fprintf(stderr, "kvm_create_irqchip failed\n");
	    goto err;
	}
    }

    for (i = 0; i < xen->vcpus; i++) {
	if (kvm_create_vcpu(xen->kvm, i) < 0) {
	    fprintf(stderr, "kvm_create_vcpu failed\n");
	    goto err;
	}
	xen->vcpu[i].id = i;
	xen->vcpu[i].vm = xen;
	need_regs(xen->vcpu + i);
	need_sregs(xen->vcpu + i);
    }
    xen->vcpu[0].online = 1;

    banner_print(xen, "setup");
    sigsetup(xen);
    if (0 != evtchn_init(xen))
	goto err;
    if (0 != xenstore_init_early(xen))
	goto err;
    console_init(xen);

    if (0 != domain_builder(xen)) {
	fprintf(stderr, "domain builder failed\n");
	goto err;
    }

    for (i = 0; i < xen->vcpus; i++) {
	if (!xen->vcpu[i].online)
	    continue;
	flush_regs(xen->vcpu + i);
	flush_sregs(xen->vcpu + i);
    }

    xenstore_init_late(xen);
    setup_cpuid(xen);

    tsc_calibrate(xen, &xen->tsc_shift, &xen->tsc_mul_frac);
    xen->boot = get_systime();
    for (i = 0; i < xen->vcpus; i++) {
	xen->vcpu[i].period = XEN_DEFAULT_PERIOD;
	setup_timer(xen->vcpu + i);
    }

    vmcore_write_headers(xen);
    pthread_mutex_lock(&xen->biglock);
    return 0;

err:
    if (xen->kvm)
	kvm_finalize(xen->kvm);
    if (-1 != xen->corefd)
	unlink(xen->corefile);
    return -1;
}

int kvmbits_vcpu_up(struct xenvcpu *vcpu)
{
    if (vcpu->thread)
	return -EINVAL;

    setup_regs(vcpu);
    flush_regs(vcpu);
    flush_sregs(vcpu);

    pthread_create(&vcpu->thread, NULL, vcpu_thread, vcpu);
    return 0;
}

int kvmbits_vcpu_down(struct xenvcpu *vcpu)
{
    /* FIXME */
    return -ENOSYS;
}
